#ifndef IST_BASE_POLYGON_MODEL
#define IST_BASE_POLYGON_MODEL


#include <vector>
#include <list>
#include <string>
#include <memory>
#include <set>
#include <boost/shared_ptr.hpp>
#include "matrix.h"


namespace ist {



struct BoundingSphere
{
	vector4 pos;
	float radius;

	BoundingSphere() : radius(0) {}
	BoundingSphere(const vector4& p, float r) : pos(p), radius(r) {}
};


struct BoundingBox
{
	vector4 min, max;

	BoundingBox(){}
	BoundingBox(const vector4 &min_, const vector4 &max_) : min(min_), max(max_) {}

	BoundingBox getAABB(const matrix44 &trans) const;
	bool overlap(const BoundingBox& bb) const;
	bool inner_point(const BoundingSphere&) const;
	bool inner_point(const vector4& pos, float radius=0) const;
	int inner_triangle(const vector4& v1, const vector4& v2, const vector4& v3) const;
};



class TriangleIndex {
public:
	TriangleIndex(int _0, int _1, int _2) : i0(_0), i1(_1), i2(_2) {}

	const int& operator [](int index) const { return v[index]; }

private:
	union {
		struct { int i0, i1, i2; };
		int v[3];
	};
};

class LineIndex {
public:
	LineIndex(int _0, int _1) {
		union_face[0]=0;
		union_face[1]=0;

		if(_0 < _1)	{ i0=_0; i1=_1; }
		else		{ i0=_1; i1=_0; }
	}

	int getKey() const { return (i0*1000000 + i1); }
	bool operator < (const LineIndex &lhs) const { return getKey() < lhs.getKey(); }
	bool operator ==(const LineIndex &lhs) const { return (i0==lhs.i0 && i1==lhs.i1); }

	void setFace(int idx, const int *f) {
		for(int i=0; i<2; ++i) {
			if(union_face[i]==0) {
				union_face[i]=f;
				face_index[i]=idx;
				break;
			}
			else if(union_face[i]==f)
				return;
		}
	}

	int getIndex(int i) const { return v[i]; }
	const int* getFace(int i) const { return union_face[i]; }
	int getFaceIndex(int i) const { return face_index[i]; }

private:
	union {
		struct { int i0, i1; };
		int v[2];
	};

	int face_index[2];
	const int *union_face[2];
};



class BasePolygonChunc
{
public:
	typedef std::vector<vector4> vertex_cont;
	typedef std::vector<vector4> normal_cont;
	typedef std::vector<vector3> vnormal_cont;
	typedef std::vector<vector2> texcoord_cont;
	typedef std::vector<int>     index_cont;
	typedef std::set<LineIndex>  line_cont;


	BasePolygonChunc();
	virtual ~BasePolygonChunc();

	virtual void clear();
	virtual void draw() const;
	virtual void setup();
	virtual void setLines();
	virtual void setNormal();
	virtual void setBoundingVolume();

	unsigned int getFaceNum()   const { return idx_array.size()/3; }
	unsigned int getVertexNum() const { return vtx_array.size(); }
	virtual const std::string& getTypeString() const;


	vertex_cont   vtx_array;
	normal_cont   nml_array; // ʖ̖@
	vnormal_cont  vnml_array; // _̖@
	texcoord_cont tex_array;
	index_cont    idx_array;
	line_cont     line_array;
	std::string   label;
	BoundingBox    bounding_box;
	BoundingSphere bounding_sphere;
	float facet;
};


class BasePolygonModel : public std::list< boost::shared_ptr<BasePolygonChunc> >
{
public:
	typedef boost::shared_ptr<BasePolygonChunc> chunc_ptr;

	BasePolygonModel() {}
	BasePolygonModel(const std::string& filename) { load(filename); }
	virtual ~BasePolygonModel() {}

	virtual bool load(const std::string& filename);
	virtual void draw() const;
	virtual void setup();

	const std::string& getFilename() const { return m_filename; }
	unsigned int getChuncNum() const { return size(); }
	unsigned int getFaceNum()  const;
	unsigned int getVertexNum()const;

	const chunc_ptr operator [](int) const;
	const chunc_ptr operator [](const std::string&) const;

	void addChunc(const chunc_ptr& p) { push_back(p); }
	virtual chunc_ptr createChunc() { return chunc_ptr(new BasePolygonChunc()); }

protected:
	std::string m_filename;

	bool loadMQO(const std::string& filename);
	bool loadOBJ(const std::string& filename);
};



// TBasePolygonChuncpclass
template<class T>
class TPolygonModel : public BasePolygonModel
{
public:
	typedef T chunc_type;

	TPolygonModel() {}
	TPolygonModel(const std::string& filename) { load(filename); }
	// TPolygonModel(const std::string& filename) : BasePolygonModel(filename) {}
	// createChunc̃I[o[[hfȂB

	chunc_ptr createChunc() { return chunc_ptr(new chunc_type()); }
};


} // ist


#endif


