#include "gl_include.h"
#include "gl_texture.h"
#include "bitmap.h"

using namespace std;

namespace ist {
namespace gl {


int TextureDimension2D::dimension() const {
	return GL_TEXTURE_2D;
}

void TextureDimension2D::assign(unsigned int name) const {
	glBindTexture(dimension(), name);
}


void TextureDimension2D::setParameter(unsigned int *tex_name) {
	glPixelStorei(GL_UNPACK_ALIGNMENT, 1);
	glGenTextures(1, tex_name);
	glBindTexture(dimension(), *tex_name);

	glTexParameteri(dimension(), GL_TEXTURE_WRAP_S, GL_REPEAT);
	glTexParameteri(dimension(), GL_TEXTURE_WRAP_T, GL_REPEAT);
	glTexParameteri(dimension(), GL_TEXTURE_MAG_FILTER, GL_LINEAR);
	glTexParameteri(dimension(), GL_TEXTURE_MIN_FILTER, GL_LINEAR);
	glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
}


template<class Dimension>
void TextureGeneratorRGBA<Dimension>::generate(unsigned int *tex_name, int width, int height, unsigned char *pixel)
{
	setParameter(tex_name);
	glTexImage2D(dimension(), 0, GL_RGBA, width, height, 0, GL_RGBA, GL_UNSIGNED_BYTE, pixel);
}

template<class Dimension>
void TextureGeneratorRGB<Dimension>::generate(unsigned int *tex_name, int width, int height, unsigned char *pixel)
{
	vector<unsigned char> data(width*height*3);
	for(unsigned int i=0; i<height; ++i) {
		for(unsigned int j=0; j<width; ++j) {
			data[i*width*3 +j*3 +0] = pixel[i*width*4 +j*4 +0];
			data[i*width*3 +j*3 +1] = pixel[i*width*4 +j*4 +1];
			data[i*width*3 +j*3 +2] = pixel[i*width*4 +j*4 +2];
		}
	}

	setParameter(tex_name);
	glTexImage2D(dimension(), 0, GL_RGB, width, height, 0, GL_RGB, GL_UNSIGNED_BYTE, &data[0]);
}

template<class Dimension>
void TextureGeneratorAlpha<Dimension>::generate(unsigned int *tex_name, int width, int height, unsigned char *pixel)
{
	vector<unsigned char> data(width*height);
	for(unsigned int i=0; i<height; ++i) {
		for(unsigned int j=0; j<width; ++j) {
			data[i*width +j]=pixel[i*width*4 +j*4 +3];
		}
	}

	setParameter(tex_name);
	glTexImage2D(dimension(), 0, GL_ALPHA, width, height, 0, GL_ALPHA, GL_UNSIGNED_BYTE, &data[0]);
}




template<class Generator> TTexture<Generator>::~TTexture() { clear(); }

template<class Generator> TTexture<Generator>::TTexture() {}
template<class Generator> TTexture<Generator>::TTexture(const char *filename) { load(filename); }

template<class Generator>
TTexture<Generator>::TTexture(const char *filename, unsigned int cell_width, unsigned int cell_height)
{
	load(filename, cell_width, cell_height);
}

template<class Generator>
TTexture<Generator>::TTexture(const TTexture<Generator> &base)
{
	*this=base;
}

template<class Generator>
TTexture<Generator>& TTexture<Generator>::operator =(const TTexture<Generator> &base)
{
	tex_name=base.tex_name;
	const_cast<TTexture<Generator>&>(base).clear();
	return *this;
}

template<class Generator>
void TTexture<Generator>::clear()
{
	if(!tex_name.empty()) {
		glDeleteTextures(tex_name.size(), &tex_name[0]);
		tex_name.clear();
	}
}

template<class Generator>
bool TTexture<Generator>::load(const char *filename)
{
	clear();

	if(filename==0) return false;

	Bitmap bmp;
	if ( !bmp.load(filename))
		return false;

	tex_name.push_back(0);
	generator.generate(&tex_name.back(), bmp.getWidth(), bmp.getHeight(), bmp[0][0].v);
	return true;
}

template<class Generator>
bool TTexture<Generator>::load(const char *filename, unsigned int cell_width, unsigned int cell_height)
{
	clear();

	if(filename==0 || cell_width<16 || cell_height<16)	// 1}X16*16̃eNX`͍Ȃ
		return false;

	Bitmap bmp;
	if ( !bmp.load(filename))
		return false;

	vector<bRGBA> cell;
	unsigned int repeat_x=bmp.getWidth()/cell_width;
	unsigned int repeat_y=bmp.getHeight()/cell_height;

	tex_name.resize(repeat_x*repeat_y);
	cell.resize(cell_width*cell_height);

	for(unsigned int i=0; i<tex_name.size(); i++) {
		for(unsigned int y=0; y<cell_height; y++) {
			for(unsigned int x=0; x<cell_width; x++) {
				cell[ y*cell_width+x ]=bmp[(i/repeat_x)*cell_height+y][(i%repeat_x)*cell_width+x];
			}
		}
		generator.generate(&tex_name[i], cell_width, cell_height, cell[0].v);
	}

	return true;
}

template<class Generator>
void TTexture<Generator>::assign(unsigned int index) const
{
	if(!tex_name.empty() && index<tex_name.size() )
		generator.assign(tex_name[index]);
}


template TTexture< TextureGeneratorRGBA<TextureDimension2D> >;
template TTexture< TextureGeneratorRGB<TextureDimension2D> >;
template TTexture< TextureGeneratorAlpha<TextureDimension2D> >;

	
}	// gl
}	// ist

