#ifndef ISTL_VECTOR_INCLUDE
#define ISTL_VECTOR_INCLUDE


#include <math.h>

namespace ist {

const float radian=3.14159f/180.0f;


struct vector2
{
	union {
		struct { float x,y; };
		float v[2];
	};

	float& operator [](int i) { return v[i]; }
	const float& operator [](int i) const { return v[i]; }

	vector2 operator - () const { return vector2(-x, -y); }
	vector2 operator * (float f) const { return vector2(x*f, y*f); }
	vector2 operator / (float f) const { return vector2(x/f, y/f); }
	void operator *=(float f) { x*=f; y*=f; }
	void operator /=(float f) { x/=f; y/=f; }
	vector2 operator + (const vector2 &a) const { return vector2(x+a.x, y+a.y); }
	vector2 operator - (const vector2 &a) const { return vector2(x-a.x, y-a.y); }
	vector2 operator * (const vector2 &a) const { return vector2(x*a.x, y*a.y); }
	vector2 operator / (const vector2 &a) const { return vector2(x/a.x, y/a.y); }
	void operator +=(const vector2 &a) { x+=a.x; y+=a.y; }
	void operator -=(const vector2 &a) { x-=a.x; y-=a.y; }
	void operator *=(const vector2 &a) { x*=a.x; y*=a.y; }
	void operator /=(const vector2 &a) { x/=a.x; y/=a.y; }
	bool operator ==(const vector2 &a) const { return x==a.x && y==a.y; }
	bool operator !=(const vector2 &a) const { return x!=a.x || y!=a.y; }


	vector2(float _x=0, float _y=0) : x(_x), y(_y) { }
	explicit vector2(const float *fv) : x(fv[0]), y(fv[1]) { }

	float square() const
	{
		return (x*x + y*y);
	}

	float norm() const
	{
		return sqrtf(x*x + y*y);
	}

	float dot(const vector2 &v) const
	{
		return (x*v.x + y*v.y);
	}

	vector2& normalize()
	{
		float l=norm();
		if(l!=0)
			*this/=l;
		return *this;
	}

	float cos(const vector2 &a) const
	{
		return acosf( dot(a) / (a.norm()*norm()) );
	}
};


struct vector3
{
	union {
		struct { float x,y,z; };
		float v[3];
	};

	float& operator [](int i) { return v[i]; }
	const float& operator [](int i) const { return v[i]; }

	vector3 operator - () const { return vector3(-x, -y, -z); }
	vector3 operator * (float f) const { return vector3(x*f, y*f, z*f); }
	vector3 operator / (float f) const { return vector3(x/f, y/f, z/f); }
	void operator *=(float f) { x*=f; y*=f; z*=f; }
	void operator /=(float f) { x/=f; y/=f; z/=f; }
	vector3 operator + (const vector3 &a) const { return vector3(x+a.x, y+a.y, z+a.z); }
	vector3 operator - (const vector3 &a) const { return vector3(x-a.x, y-a.y, z-a.z); }
	vector3 operator * (const vector3 &a) const { return vector3(x*a.x, y*a.y, z*a.z); }
	vector3 operator / (const vector3 &a) const { return vector3(x/a.x, y/a.y, z/a.z); }
	void operator +=(const vector3 &a) { x+=a.x; y+=a.y; z+=a.z; }
	void operator -=(const vector3 &a) { x-=a.x; y-=a.y; z-=a.z; }
	void operator *=(const vector3 &a) { x*=a.x; y*=a.y; z*=a.z; }
	void operator /=(const vector3 &a) { x/=a.x; y/=a.y; z/=a.z; }
	bool operator ==(const vector3 &a) const { return x==a.x && y==a.y && z==a.z; }
	bool operator !=(const vector3 &a) const { return x!=a.x || y!=a.y || z!=a.z; }


	vector3(float _x=0, float _y=0, float _z=0) : x(_x), y(_y), z(_z) { }
	explicit vector3(const float *fv) : x(fv[0]), y(fv[1]), z(fv[2]) { }

	float square() const
	{
		return (x*x + y*y + z*z);
	}

	float norm() const
	{
		return sqrtf(x*x + y*y + z*z);
	}

	float dot(const vector3 &v) const
	{
		return (x*v.x + y*v.y + z*v.z);
	}

	vector3& normalize()
	{
		float l=norm();
		if(l!=0)
			*this/=l;
		return *this;
	}

	vector3& cross(const vector3 &a, const vector3 &b)
	{
		x = a.y*b.z - a.z*b.y;
		y = a.z*b.x - a.x*b.z;
		z = a.x*b.y - a.y*b.x;
		return *this;
	}

	float cos(const vector3 &a) const
	{
		return acosf( dot(a) / (a.norm()*norm()) );
	}
};


struct vector4
{
	union {
		struct { float x,y,z,w; };
		float v[4];
	};

	float& operator [](int i) { return v[i]; }
	const float& operator [](int i) const { return v[i]; }

	vector4 operator - () const { return vector4(-x, -y, -z); }
	vector4 operator * (float f) const { return vector4(x*f, y*f, z*f); }
	vector4 operator / (float f) const { return vector4(x/f, y/f, z/f); }
	void operator *=(float f) { x*=f; y*=f; z*=f; }
	void operator /=(float f) { x/=f; y/=f; z/=f; }
	vector4 operator + (const vector4 &a) const { return vector4(x+a.x, y+a.y, z+a.z); }
	vector4 operator - (const vector4 &a) const { return vector4(x-a.x, y-a.y, z-a.z); }
	vector4 operator * (const vector4 &a) const { return vector4(x*a.x, y*a.y, z*a.z); }
	vector4 operator / (const vector4 &a) const { return vector4(x/a.x, y/a.y, z/a.z); }
	void operator +=(const vector4 &a) { x+=a.x; y+=a.y; z+=a.z; }
	void operator -=(const vector4 &a) { x-=a.x; y-=a.y; z-=a.z; }
	void operator *=(const vector4 &a) { x*=a.x; y*=a.y; z*=a.z; }
	void operator /=(const vector4 &a) { x/=a.x; y/=a.y; z/=a.z; }
	bool operator ==(const vector4 &a) const { return x==a.x && y==a.y && z==a.z && w==a.w; }
	bool operator !=(const vector4 &a) const { return x!=a.x || y!=a.y || z!=a.z || w!=a.w; }


	vector4(float _x=0, float _y=0, float _z=0, float _w=1) : x(_x), y(_y), z(_z), w(_w) { }
	explicit vector4(const float *fv) : x(fv[0]), y(fv[1]), z(fv[2]), w(fv[3]) { }
	vector4(const vector3& v3, float _w=1) : x(v3.x), y(v3.y), z(v3.z), w(_w) {}

	vector3 to_vector3() const { return vector3(v); }

	float square() const
	{
		return (x*x + y*y + z*z);
	}

	float norm() const
	{
		return sqrtf(x*x + y*y + z*z);
	}

	float dot(const vector4 &v) const
	{
		return (x*v.x + y*v.y + z*v.z);
	}

	float dot4(const vector4 &v) const
	{
		return (x*v.x + y*v.y + z*v.z + w*v.w);
	}

	vector4& normalize()
	{
		float l=norm();
		if(l!=0)
			*this/=l;
		return *this;
	}

	vector4& cross(const vector4 &a, const vector4 &b)
	{
		x = a.y*b.z - a.z*b.y;
		y = a.z*b.x - a.x*b.z;
		z = a.x*b.y - a.y*b.x;
		return *this;
	}

	float cos(const vector4 &a) const
	{
		return acosf( dot(a) / (a.norm()*norm()) );
	}
};

} // ist

#endif

